require("defines")
require("util")
require("helpers.gui_helpers")

local data={}
local classes = {}
--local _debug=false

local function debugLog(message)
  if _debug then -- set for debug
    game.player.print(tostring(message))
  end
end

local function ttostring(tab)
  local res=""
  if type(tab)=="table" then
    res="{"
    for k,v in pairs(tab) do
      res=res.." "..tostring(k).." = "..ttostring(v)..","
    end
    res=res.."}"
  else
    res=tostring(tab)
  end
  return res
end

--Rail Switch
classes["rail-switch-lamp"]={
	onPlace=function(entity)
		debugLog("Placed switch")
		local rail = entity.surface.find_entity("straight-rail",entity.position)
		return {entity=entity,rail=rail,active=(rail~=nil)}
	end,
  onDestroy=function(entity)
	debugLog("Removed switch")
	if entity.rail and entity.rail.valid then
		entity.rail.destroy()
	end
  end,
  onTick=function(entity)
    local r=entity.entity.get_circuit_condition(1).fulfilled and entity.entity.energy > 0.5
    if r~=entity.active then
      if r then
		entity.rail=entity.entity.surface.find_entity("straight-rail",entity.entity.position) or entity.entity.surface.create_entity({name="straight-rail",position=entity.entity.position,force=entity.entity.force,direction=entity.direction})
      else
        if entity.rail and entity.rail.valid then
          entity.direction=entity.rail.direction
          entity.rail.destroy()
		  entity.rail=nil
        else
          local rail = entity.entity.surface.find_entity("straight-rail",entity.entity.position)
          if rail then
            entity.direction=rail.direction
            rail.destroy()
          else
            entity.entity.die()
          end
        end
      end
      entity.active=r
    end
  end
}


local function getArea(pos)
  return {{pos.x-1,pos.y-1},{pos.x+1,pos.y+1}}
end

--Rail Signal Reader
classes["rail-signal-detector"]={
  onPlace=function(entity) return {entity=entity} end,
  onDestroy=function() end,
  onTick=function(k)
    local sum=0
    for i1,k1 in ipairs(k.entity.surface.find_entities_filtered({area=getArea(k.entity.position),type="rail-signal"})) do
      sum=sum+k1.signal_state
    end
    for i1,k1 in ipairs(k.entity.surface.find_entities_filtered({area=getArea(k.entity.position),type="rail-chain-signal"})) do
      sum=sum+k1.signal_state
    end
    local r=k.entity.get_circuit_condition(1)
    for i1,k1 in pairs(r.parameters) do
      k1.count=sum
    end
    k.entity.set_circuit_condition(1,r)
  end
}

local function getTrain(entity)
  return entity.train
end

local function OpenSCGUI(playerIndex,k,i)
  if k.gui[playerIndex] ~= nil then
    return
  end
  debugLog("Open Gui")
  GUI.PushLeftSection(playerIndex)
  k.gui[playerIndex] = GUI.PushParent(GUI.Frame("station_controller_"..i, "Set Destination Station", GUI.HORIZONTAL))
  GUI.PushParent(GUI.Flow("flow", GUI.VERTICAL))
  GUI.TextField("input", k.station).text = k.station
  GUI.PopAll()
end

local function CloseSCGUI(playerIndex,k,i)
  if k.gui[playerIndex]==nil then
    return
  end
  debugLog("Close Gui")
  k.gui[playerIndex].destroy()
  k.gui[playerIndex]=nil
end

classes["rail-station-controller"]={
onPlace=function(entity)
  return {entity=entity,gui={},station=""}
end,
onDestroy=function(entity)
	for k,v in pairs(entity.gui) do
		v.destroy()
	end
end,
onTick=function(obj,index)
  for playerIndex = 1, #game.players do
    if util.distance(game.players[playerIndex].position, obj.entity.position) < 2 then
      OpenSCGUI(playerIndex,obj,index)
    else
      CloseSCGUI(playerIndex,obj,index)
    end
    if obj.gui[playerIndex] then
      obj.station=obj.gui[playerIndex].flow.input.text
    end
  end

  if obj.entity.get_circuit_condition(1).fulfilled and obj.entity.energy > 0.5 then
	for i,k in ipairs(obj.entity.surface.find_entities(getArea(obj.entity.position))) do
	  local ok,train=pcall(getTrain,k)
	  if ok then
		for i,k in ipairs(train.schedule.records) do
			if k.station == obj.station then
				local sch=train.schedule
                                if sch.current ~= i then
                                    if not train.manual_mode then
                                        train.manual_mode = true
                                        train.manual_mode = false
                                    end
                                    sch.current=i
                                    train.schedule=sch
                                    break
                                end
			end
		end
		break
	  end
	end
  end
end
}

--Rail Train Detector
classes["rail-train-detector"]={
  onPlace=function(entity) return {entity=entity} end,
  onDestroy=function() end,
  onTick=function(k)
    local res=0
    for i,k in ipairs(k.entity.surface.find_entities(getArea(k.entity.position))) do
	  local ok,train=pcall(getTrain,k)
	  if ok then
		res=1
		break
	  end
    end
    local r=k.entity.get_circuit_condition(1)
    for i1,k1 in pairs(r.parameters) do
      k1.count=res
    end
    k.entity.set_circuit_condition(1,r)
  end
}

--Cargo Inventory Reader
classes["rail-cargo-reader"]={
  onPlace=function(entity) return {entity=entity} end,
  onDestroy=function() end,
  onTick=function(k)
	local cargos = (k.entity.surface.find_entities_filtered({area=getArea(k.entity.position),type="cargo-wagon"}))
	
	local out=k.entity.get_circuit_condition(1)
	local inv = {}
	for i=1,#cargos do
		for k,v in pairs(cargos[i].get_inventory(1).get_contents()) do
			inv[k] = (inv[k] or 0) + v
		end
	end

	local i=1
	for k,v in pairs(inv) do
		out.parameters[i]={signal={type = "item",name = k},count=v,index = i}
		i=i+1
		if i>15 then break end
	end
	for c=i,15 do 
		out.parameters[i]={signal={type = "item",name = nil},count = 1,index = i}
	end
	k.entity.set_circuit_condition(1,out)
  end
}


local function OpenMCGUI(playerIndex,k,i)
  if k.gui[playerIndex] ~= nil then
    return
  end
  debugLog("Open Gui")
  GUI.PushLeftSection(playerIndex)
  k.gui[playerIndex] = {GUI.PushParent(GUI.Frame("mode_controller_"..i, "When condition then", GUI.HORIZONTAL))}
  GUI.PushParent(GUI.Flow("flow", GUI.VERTICAL))
  k.gui[playerIndex].checkbox=GUI.CheckboxNOCLB( "setmode", "Set mode to", k.setmode )
  GUI.PopAll()
end

local function CloseMCGUI(playerIndex,k,i)
  if k.gui[playerIndex]==nil then
    return
  end
  debugLog("Close Gui")
  k.gui[playerIndex][1].destroy()
  k.gui[playerIndex]=nil
end

classes["rail-mode-controller"]={
onPlace=function(entity)
  return {entity=entity,gui={},setmode=false,checkbox=nil}
end,
onDestroy=function(entity)
	for k,v in pairs(entity.gui) do
		v.destroy()
	end
end,
onTick=function(obj,index)
  for playerIndex = 1, #game.players do
    if util.distance(game.players[playerIndex].position, obj.entity.position) < 2 then
      OpenMCGUI(playerIndex,obj,index)
    else
      CloseMCGUI(playerIndex,obj,index)
    end
    if obj.gui[playerIndex] then
      obj.setmode = obj.gui[playerIndex].checkbox.state
    end
  end

  if obj.entity.get_circuit_condition(1).fulfilled and obj.entity.energy > 0.5 then
	for i,k in ipairs(obj.entity.surface.find_entities(getArea(obj.entity.position))) do
	  local ok,train=pcall(getTrain,k)
	  if ok then
		train.manual_mode = obj.setmode
		break
	  end
	end
  end
end
}

--Cargo Inventory Reader
classes["rail-cargo-reader2"]={
  onPlace=function(entity) return {entity=entity} end,
  onDestroy=function() end,
  onTick=function(k)
        local cargos={}
	for i,k in ipairs(k.entity.surface.find_entities(getArea(k.entity.position))) do
	  local ok,train=pcall(getTrain,k)
	  if ok then
		cargos=train.carriages
		break
	  end
        end
	local out=k.entity.get_circuit_condition(1)
	local inv = {}
	for i=1,#cargos do
            if cargos[i].type == "cargo-wagon" then
		for k,v in pairs(cargos[i].get_inventory(1).get_contents()) do
			inv[k] = (inv[k] or 0) + v
		end
            end
	end

	local i=1
	for k,v in pairs(inv) do
		out.parameters[i]={signal={type = "item",name = k},count=v,index = i}
		i=i+1
		if i>15 then break end
	end
	for c=i,15 do 
		out.parameters[i]={signal={type = "item",name = nil},count = 1,index = i}
	end
	k.entity.set_circuit_condition(1,out)
  end
}


local function entityBuilt(event)
  if classes[event.created_entity.name]~= nil then
	local tab = data[event.created_entity.name]
    table.insert(tab,classes[event.created_entity.name].onPlace(event.created_entity))
	data[event.created_entity.name] = tab
  end
end

local function entityRemoved(event)
  if classes[event.entity.name]~= nil then
	for k, v in ipairs(data[event.entity.name]) do
      if v.entity==event.entity then
		local tab = data[event.entity.name]
        table.remove(tab,k)
		classes[event.entity.name].onDestroy(v)
		data[event.entity.name] = tab
        break
      end
    end
  end
end

local function onTick()
	for k, v in pairs(classes) do
		for q, i in pairs(data[k]) do
			if i.entity.valid then
				v.onTick(i,q)
			end
		end
	end
end

local function onLoad()
  data = global.railLogicSystem or {}
  for k,v in pairs(classes) do
	data[k]=data[k] or {}
  end
end

game.on_init(onLoad)
game.on_load(onLoad)
game.on_save(function()
    global.railLogicSystem = data
end)

game.on_event(defines.events.on_tick, onTick)

game.on_event(defines.events.on_built_entity, entityBuilt)
game.on_event(defines.events.on_robot_built_entity, entityBuilt)

game.on_event(defines.events.on_preplayer_mined_item, entityRemoved)
game.on_event(defines.events.on_robot_pre_mined, entityRemoved)
game.on_event(defines.events.on_entity_died, entityRemoved)